import unittest
from unittest.mock import patch, Mock
from lordoftherings import LordOfTheRings
import requests

class TestLordOfTheRings(unittest.TestCase):
    def setUp(self):
        self.auth_token = 'YOUR API TOKEN'
        self.api = LordOfTheRings(self.auth_token)

    # Test the authorization header being set properly
    def test_authorization_header(self):
        api = LordOfTheRings('test-token')
        self.assertEqual(api.session.headers['Authorization'], 'Bearer test-token')

    # Testing /movie
    @patch('lordoftherings.lordoftherings.LordOfTheRings.request')
    def test_get_all_movies(self, mock_request):
        self.api.movies().get_all().fetch()
        mock_request.assert_called_once_with('GET', 'movie', params={})

    # Testing /movie/{id}
    @patch('lordoftherings.lordoftherings.LordOfTheRings.request')
    def test_get_movie(self, mock_request):
        movie_id = '5cd95395de30eff6ebccde56'
        self.api.movies(movie_id).get()
        mock_request.assert_called_once_with('GET', f'movie/{movie_id}')

    # Testing /movie/{id}/quote
    @patch('lordoftherings.lordoftherings.LordOfTheRings.request')
    def test_get_movie_quotes(self, mock_request):
        movie_id = '5cd95395de30eff6ebccde56'
        self.api.movies(movie_id).quotes.get_all().fetch()
        mock_request.assert_called_once_with('GET', f'movie/{movie_id}/quote', params={})

    # Testing /quote
    @patch('lordoftherings.lordoftherings.LordOfTheRings.request')
    def test_get_all_quotes(self, mock_request):
        self.api.quotes().get_all().fetch()
        mock_request.assert_called_once_with('GET', 'quote', params={})

    # Testing /quote/{id}
    @patch('lordoftherings.lordoftherings.LordOfTheRings.request')
    def test_get_quote(self, mock_request):
        quote_id = '5cd96e05de30eff6ebccfb2b'
        self.api.quotes(quote_id).get()
        mock_request.assert_called_once_with('GET', f'quote/{quote_id}')

    # Testing pagination when just the limit() function is used
    @patch.object(requests.Session, 'request')
    def test_limit_pagination(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {'docs': []})
        limit = 5
        self.api.movies().get_all().limit(limit).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'limit': limit})

    # Testing pagination when just the page() function is used
    @patch.object(requests.Session, 'request')
    def test_page_pagination(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {'docs': []})
        page = 2
        self.api.movies().get_all().page(page).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'page': page})

    # Testing pagination when just the offset() function is used
    @patch.object(requests.Session, 'request')
    def test_offset_pagination(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {'docs': []})
        offset = 10
        self.api.movies().get_all().offset(offset).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'offset': offset})

    # Testing pagination when all functions for pagination are used together
    @patch.object(requests.Session, 'request')
    def test_combined_pagination(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {'docs': []})
        limit = 5
        page = 2
        offset = 10
        self.api.movies().get_all().limit(limit).page(page).offset(offset).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'limit': limit, 'page': page, 'offset': offset})

    # Testing filtering match
    @patch.object(requests.Session, 'request')
    def test_filtering_match(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {})
        self.api.movies().get_all().filter_by(name="The Lord of the Rings Series").fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'name': 'The Lord of the Rings Series'})

    # Testing filtering negate match
    @patch.object(requests.Session, 'request')
    def test_filtering_negate_match(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {})
        self.api.movies().get_all().filter_by(name=('not', 'Frodo')).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'name!': 'Frodo'})

    # Testing filtering includes
    @patch.object(requests.Session, 'request')
    def test_filtering_includes(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {})
        self.api.movies().get_all().filter_by(name=['The Lord of the Rings Series', 'The Desolation of Smaug']).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'name': 'The Lord of the Rings Series,The Desolation of Smaug'})
    
    # Testing filtering excludes
    @patch.object(requests.Session, 'request')
    def test_filtering_excludes(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {})
        self.api.movies().get_all().filter_by(name=[('not', 'The Lord of the Rings Series'), ('not', 'The Desolation of Smaug')]).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'name!': 'The Lord of the Rings Series,The Desolation of Smaug'})

    # Testing filtering less than
    @patch.object(requests.Session, 'request')
    def test_filtering_less_than(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {})
        self.api.movies().get_all().filter_by(budgetInMillions=('<', 100)).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'budgetInMillions<': 100})

    # Testing filtering greater than
    @patch.object(requests.Session, 'request')
    def test_filtering_greater_than(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {})
        self.api.movies().get_all().filter_by(runtimeInMinutes=('>', 160)).fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'runtimeInMinutes>': 160})

    # Testing sorting with only the sort_by() function
    @patch.object(requests.Session, 'request')
    def test_sorting(self, mock_request):
        mock_request.return_value = Mock(ok=True, status_code=200, json=lambda: {})
        self.api.movies().get_all().sort_by('name:desc').fetch()
        mock_request.assert_called_once_with('GET', f'{self.api.base_url}movie', params={'sort': 'name:desc'})

    # Testing status code 429 to check that rate limits have been handled
    @patch.object(requests.Session, 'request', side_effect=[Mock(ok=False, status_code=429, json=lambda: {})]*4)
    def test_rate_limit_error_handling(self, mock_request):
        with self.assertRaises(Exception) as context:
            self.api.movies('5cd95395de30eff6ebccde56').get()
        self.assertTrue('Max retries exceeded' in str(context.exception))
        self.assertEqual(mock_request.call_count, self.api.max_retries)

    # Testing status code 500 for server errors
    @patch.object(requests.Session, 'request', side_effect=[Mock(ok=False, status_code=500, json=lambda: {})]*4)
    def test_server_error_handling(self, mock_request):
        with self.assertRaises(Exception) as context:
            self.api.movies('5cd95395de30eff6ebccde56').get()
        self.assertTrue('Max retries exceeded' in str(context.exception))
        self.assertEqual(mock_request.call_count, self.api.max_retries)

    # Testing the status code 400 for client errors when calling the api
    @patch.object(requests.Session, 'request', return_value=Mock(ok=False, status_code=400))
    def test_client_error_handling(self, mock_request):
        with self.assertRaises(Exception) as context:
            self.api.movies('5cd95395de30eff6ebccde56').get()
        self.assertTrue('Request failed with status 400' in str(context.exception))

if __name__ == '__main__':
    unittest.main()



import requests
import time

class APIResource:
    """Class that is used to represent something that will be returned from the api

    Attributes
    ----------
    api : LordOfTheRings
        The instance of the LordOfTheRings class that this resource will use to make requests
    endpoint : str
        The endpoint associated with this API resource
    id : str, optional
        The id of the resource you wish to get from the api.

    """
    def __init__(self, api, endpoint, id=None):
        """Initialize an APIResource instance.

        Parameters
        ----------
        api : LordOfTheRings
            The instance of the LordOfTheRings class that this resource will use to make requests.
        endpoint : str
            The endpoint associated with this API resource.
        id : str, optional
            The unique identifier of this resource (default is None).

        """
        self.api = api
        self.endpoint = endpoint
        self.id = id

    def get_all(self):
        """Fetch all resources of this type

        Returns
        -------
        ChainableMethods
            An instance of ChainableMethods to allow for option chaining like filtering, sorting, or pagination.

        """
        return ChainableMethods(self)

    def get(self):
        """Fetch a single resource by the id of the resource

        Returns
        -------
        dict
            The JSON response with the data from the resource that was queried

        Raises
        ------
        Exception
            If the ID of the resource has not been set.

        """
        if self.id:
            return self.api.request('GET', f'{self.endpoint}/{self.id}')
        else:
            raise Exception('Resource ID not provided')
        
class ChainableMethods:
    """Class to allow chainable methods on a resource.

    Attributes
    ----------
    resource : APIResource
        The resource that will have method chaining.
    params : dict
        A dictionary of parameters for the API request.

    """
    def __init__(self, resource):
        """Initialize a ChainableMethods instance.

        Parameters
        ----------
        resource : APIResource
            The resource on which the methods should be chained.

        """
        self.resource = resource
        self.params = {}

    def limit(self, limit):
        """Set the limit parameter for the request, limitting the amount of records returned by the api

        Parameters
        ----------
        limit : int
            The maximum number of results to return.

        Returns
        -------
        ChainableMethods
            The current instance to allow method chaining.

        """
        self.params['limit'] = limit
        return self

    def page(self, page):
        """Set the page parameter for the request, specifying which page of results should be returned

        Parameters
        ----------
        page : int
            The page of results to return.

        Returns
        -------
        ChainableMethods
            The current instance to allow method chaining.

        """
        self.params['page'] = page
        return self

    def offset(self, offset):
        """Set the offset parameter for the request.

        Parameters
        ----------
        offset : int
            The number of results to skip before starting to fetch.

        Returns
        -------
        ChainableMethods
            The current instance to allow method chaining.

        """
        self.params['offset'] = offset
        return self

    def filter_by(self, **filters):
        """Set filter parameters for the request.

        Parameters
        ----------
        **filters : dict
            The filters to apply to the request.

        Returns
        -------
        ChainableMethods
            The current instance to allow method chaining.

        """
        for key, value in filters.items():
            if isinstance(value, str):
                # match
                self.params[key] = value
            elif isinstance(value, tuple):
                if value[0] == 'not':
                    # negate match
                    self.params[f"{key}!"] = value[1]
                elif value[0] in ('<', '>'):
                    # less than, greater than or equal to
                    self.params[f"{key}{value[0]}"] = value[1]
            elif isinstance(value, list):
                if all(isinstance(v, str) for v in value):
                    # include
                    self.params[key] = ','.join(value)
                elif all(isinstance(v, tuple) and v[0] == 'not' for v in value):
                    # exclude
                    self.params[f"{key}!"] = ','.join(v[1] for v in value)
        return self
    
    def sort_by(self, sort_by):
        """Set the sort parameter for the request.

        Parameters
        ----------
        sort_by : str
            The field by which to sort the results. For example, name:asc

        Returns
        -------
        ChainableMethods
            The current instance to allow method chaining.

        """
        self.params['sort'] = sort_by
        return self

    def fetch(self):
        """Make the API request and return the response.

        Returns
        -------
        dict
            The JSON response from the server.

        """
        return self.resource.api.request('GET', self.resource.endpoint, params=self.params)

class Movie(APIResource):
    """A class representing a movie resource.
    
    Attributes
    ----------
    quotes : Quotable
        A Quotable instance composed with the movie, allowing the user to get quotes for a particular movie
    """
    def __init__(self, api, id=None):
        """Initialize a Movie instance.

        Parameters
        ----------
        api : LordOfTheRings
            The main API instance.
        id : str, optional
            The ID of the movie. 
        """
        endpoint = 'movie'
        super().__init__(api, endpoint, id)
        self.quotes = Quotable(api, endpoint, self.id)

class Quote(APIResource):
    """A class representing a quote resource."""
    def __init__(self, api, id=None):
        """Initialize a Quote instance.

        Parameters
        ----------
        api : LordOfTheRings
            The main API instance.
        id : str, optional
            The ID of the quote.
        """
        endpoint = 'quote'
        super().__init__(api, endpoint, id)

class Quotable(APIResource):
    """A class representing a quotable resource, that can have quotes in the form of {resource}/{id}/quote"""
    def __init__(self, api, parent_endpoint, parent_id):
        """Initialize a Quotable instance.

        Parameters
        ----------
        api : LordOfTheRings
            The main API instance.
        parent_endpoint : str
            The endpoint of the parent resource.
        parent_id : str
            The ID of the parent resource.
        """
        super().__init__(api, f'{parent_endpoint}/{parent_id}/quote')

class LordOfTheRings:
    """The client for interfacing with the Lord of The Rings API.

    Attributes
    ----------
    base_url : str
        The base URL for the API.
    auth_token : str
        The authentication token for the API.
    session : requests.Session
        The session for making HTTP requests.
    max_retries : int
        The maximum number of retries when a request fails.
    failure_delay : int
        The delay (in seconds) between retries.
    """
    def __init__(self, auth_token, max_retries=3, failure_delay=1):
        """Initialize a LordOfTheRings client instance.

        Parameters
        ----------
        auth_token : str
            The authentication token for the API.
        max_retries : int, optional
            The maximum number of retries when a request fails.
        failure_delay : int, optional
            The delay (in seconds) between retries.
        """
        self.base_url = 'https://the-one-api.dev/v2/'
        self.auth_token = auth_token
        self.session = requests.Session()
        self.max_retries = max_retries
        self.failure_delay = failure_delay
        if auth_token:
            self.session.headers.update({'Authorization': f'Bearer {auth_token}'})

    def request(self, method, endpoint, **kwargs):
        """Send a request to the API.

        Parameters
        ----------
        method : str
            The HTTP method ('GET', 'POST', etc.).
        endpoint : str
            The API endpoint (excluding the base URL).
        **kwargs
            Any other parameters to pass to requests.Session.request.

        Returns
        -------
        dict
            The JSON response from the API.

        Raises
        ------
        Exception
            If the request fails after max_retries.
        """
        url = f'{self.base_url.rstrip("/")}/{endpoint.lstrip("/")}'
        for _ in range(self.max_retries):
            response = self.session.request(method, url, **kwargs)
            if response.status_code == 429:  # Rate limit error
                print("Rate limit reached. Retrying...")
                time.sleep(self.failure_delay)
                continue
            elif response.status_code >= 500: # Server error
                print("Server error. Retrying...")
                time.sleep(self.failure_delay)
                continue
            elif response.status_code >= 400: # Client error
                raise Exception(f'Request failed with status {response.status_code}')
            else:
                return response.json()
        raise Exception('Max retries exceeded')
    
    def movies(self, id=None):
        """Get a movie resource.

        Parameters
        ----------
        id : str, optional
            The ID of the movie.

        Returns
        -------
        Movie
            A Movie instance.
        """
        return Movie(self, id)

    def quotes(self, id=None):
        """Get a quote resource.

        Parameters
        ----------
        id : str, optional
            The ID of the quote.

        Returns
        -------
        Quote
            A Quote instance.
        """
        return Quote(self, id)


from lordoftherings import LordOfTheRings
# Add your api token to test
api = LordOfTheRings('YOUR API TOKEN')

# Uncomment and run the uses you wish to test. Due to rate limits it is not recommended that you run all at the same time

# /movie
# all_movies = api.movies().get_all().fetch()

# /movie with filters
# match = api.movies().get_all().filter_by(name='Gandalf').fetch()  # match
# negate_match = api.movies().get_all().filter_by(name=('not', 'Frodo')).fetch()  # negate match
# include = api.movies().get_all().filter_by(name=['The Lord of the Rings Series', 'The Desolation of Smaug']).fetch()  # include
# exclude = api.movies().get_all().filter_by(name=[('not', 'The Lord of the Rings Series'), ('not', 'The Desolation of Smaug')]).fetch()  # exclude
# less_than = api.movies().get_all().filter_by(budgetInMillions=('<', 100)).fetch()  # less than
# greater_than = api.movies().get_all().filter_by(runtimeInMinutes=('>', 160)).fetch()  # greater than or equal to

# /movie with sorting
# sorted_movies = api.movies().get_all().sort_by("name:asc").fetch()

# /movie with sorting and filtering
# sorted_filtered = api.movies().get_all().sort_by("name:asc").filter_by().fetch()

# /movie with pagination
# paged_movies_1 = api.movies().get_all().limit(2).fetch()
# paged_movies_2 = api.movies().get_all().limit(3).page(2).fetch()
# paged_movies_3 = api.movies().get_all().limit(1).page(1).offset(1).fetch()

# /movie/5cd95395de30eff6ebccde56
# movie = api.movies("5cd95395de30eff6ebccde56").get()

# /movie/5cd95395de30eff6ebccde56/quote
# movie_quotes = api.movies("5cd95395de30eff6ebccde5b").quotes.get_all().fetch()

# /movie/5cd95395de30eff6ebccde56/quote with sorting
# movie_quotes_sorted = api.movies("5cd95395de30eff6ebccde5b").quotes.get_all().sort_by("dialog:asc").fetch()

# /quote
# quotes = api.quotes().get_all().fetch()

# /quote with sorting
# sorted_quotes = api.quotes().get_all().sort_by('dialog:desc').fetch()

# /quote/5cd96e05de30eff6ebccebc9
# single_quote = api.quotes("5cd96e05de30eff6ebccebc9").get()



from setuptools import setup, find_packages

setup(
    name="AaronBlaser-SDK",
    version="1.0",
    author="Aaron Blaser",
    author_email="ablaser5@comcast.net",
    description="SDK for the lord of the rings API",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ablaser5/AaronBlaser-SDK",
    packages=find_packages(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    install_requires=[
        "certifi==2023.5.7",
        "charset-normalizer==3.1.0",
        "idna==3.4",
        "requests==2.31.0",
        "urllib3==2.0.2",
    ],
)